// -------------------------------------------------------------------------------------
//
// RobMotors.CS                                                   Ulrich Mller 16.06.05
//                                                                PCS1700   C# Final SP2
// Klassen zum Betrieb eines oder mehrerer RobMotoren
//
// -------------------------------------------------------------------------------------

using System;
using System.Threading;

namespace FishDevices40 {
	
	// --- RobMotor : Betrieb eines einzelnen RobMotors -----------------------

	public class RobMotor : DualOutput {

		private int MaxPos;
		private int StartPos;
		private int DestPos;
		private int ActPos;

		public delegate void Position(object sender, int Pos, bool Ready);
		public event         Position ChangedPosition;
		public event         Position FinalPosition;

		// --- Constructor ----------------------------------------------------------

		public RobMotor(FishDevices fd, Out MotNr, int MaxPosition) 
			: this(fd, true, Convert.ToInt32(MotNr), MaxPosition) {}
		public RobMotor(FishDevices fd, int MotNr, int MaxPosition) 
			: this(fd, true, MotNr, MaxPosition) {}
		public RobMotor(FishDevices fd, bool WithEvents, Out MotNr, int MaxPosition) 
			: this(fd, WithEvents, Convert.ToInt32(MotNr), MaxPosition) {}
		public RobMotor(FishDevices fd, bool WithEvents, int MotNr, int MaxPosition) 
			: base(fd, MotNr) {
			this.MaxPos   = MaxPosition;
			this.StartPos = 9999;
			this.DestPos  = 0;
			base.threadInterval = 11;
			base.WithEvents = WithEvents;  // --- Basis DualOutput hat keine Events
			if(base.WithEvents) {
				base.deviceThread = new Thread(new ThreadStart(EventLoop));
				base.deviceThread.IsBackground = true;
			}
		}

		// --- EventLoop : Bestimmen der aktuellen Position ---------------
		
		private void EventLoop() {
			bool Ready;
			while (!fd.NotHalt && !fd.ESC) {
				int ac = fd.GetCounter(this.ImpulseSwitch);
				if((uint)fd.GetInputs() == FishBase.rbFehler) 
					throw(new FishDevException("InterfaceProblem.RobMotor"));
				Ready = false;
				if(StartPos < DestPos) {
					ActPos = DestPos - ac;
					if(ac == 0) {
						Ready    = true;
						StartPos = DestPos;
					}
				}
				else {
					ActPos = DestPos + ac;
					if(ac == 0) {
						Ready    = true;
						StartPos = DestPos;
					}
				}
				if(Ready) base.stateValue = 0;
				if(ChangedPosition != null) ChangedPosition(this, ActPos, Ready);
				if(FinalPosition != null && Ready) FinalPosition(this, ActPos, Ready);
				Thread.Sleep(base.threadInterval);
			}
		}

		// --- Eigenschaften ----------------

		public int MotorNumber {
			get{return base.DeviceNumber;}
		}
		public int EndSwitch {
			get{return (base.DeviceNumber-1)*2+1;}
		}
		public int ImpulseSwitch {
			get {return (base.DeviceNumber-1)*2+2;}
		}
		public int MaxPosition {
			get {return MaxPos;}
		}
		public int CurrentPosition {            // --- in Ruhe 
			get {return StartPos;}
			set {StartPos = value;}
		}
		public int ActualPosition {             // --- laufende Pos.
			get {return ActPos;}
		}

		// --- Methoden ------------------------------

		public void DriveHome() {
			DestPos         = 0;
			base.stateValue = 1;
			try {
				fd.SetMotor(base.DeviceNumber, stateValue, 15, 9999);
			}
			catch(FishFaceException eft) {
				if(!fd.NotHalt)
					throw(new FishDevException(eft.Message + ".RobMotor.DriveHome"));}
		}

		public void DriveDelta(int Inc) {
			this.DriveTo(StartPos + Inc);
		}

		public void DriveTo(int Pos) {
			DestPos = Pos;
			if(DestPos > MaxPos) DestPos = MaxPos;
			try {
				if(DestPos > StartPos) {
					base.stateValue = 2;
					fd.SetMotor(base.DeviceNumber, base.stateValue, 15, DestPos-StartPos);
				}
				else {
					base.stateValue = 1;
					int ct = StartPos - DestPos;
					if(ct > 0) fd.SetMotor(base.DeviceNumber, base.stateValue, 15, ct); 
				}
			}
			catch(FishFaceException eft){
				if(!fd.NotHalt)
					throw(new FishDevException(eft.Message + ".RobMotor.DriveTo"));}
		}

		public void WaitForDone() {
			try {
				fd.WaitForMotor(0, base.DeviceNumber);
			}
			catch(FishFaceException eft) {
				if(!fd.NotHalt) 
					throw(new FishDevException(eft.Message + ".RobMotor.WaitForDone"));}
			StartPos        = DestPos;
			base.stateValue = 0;
		}
	}

	// --- RobMotors : Simultaner Betrieb mehrere RobMotoren ---------------

	public class RobMotors : DeviceBase {

		private RobMotor[] RMotList;
		private int[]      PosList;
		private int        anzMot;
		private int        anzPos;
		private int        stateValue;

		public delegate void PositionS(object sender, int[] Pos, bool Ready);
		public event         PositionS ChangedPosition;
		public event         PositionS FinalPosition;

		// --- Constructor ----------------------------------------------------------

		public RobMotors(FishDevices fd, params RobMotor[] RMotList)
			: this(fd, true, RMotList) {}
		public RobMotors(FishDevices fd, bool WithEvents, params RobMotor[] RMotList) 
			: base(fd, false, 103) {
			fd.AddDevice(this);
			this.anzMot = RMotList.Length;
			this.anzPos = 0;
			this.stateValue = 0;
			this.RMotList = new RobMotor[anzMot];
			for(int i = 0; i < this.anzMot; i++) {
				this.RMotList[i] = RMotList[i];
				if(WithEvents) 
					this.RMotList[i].ChangedPosition += new RobMotor.Position(EventTrigger);
			}
			PosList = new int[anzMot];
		}

		// --- Ereignis-Behandlung ----------------------------------------

		private void EventTrigger(object sender, int Pos, bool Ready) {
			Ready = true;
			for(int i = 0; i < anzMot; i++) {
				PosList[i] = RMotList[i].ActualPosition;
				if(RMotList[i].State != 0) Ready = false;
			}
			if(ChangedPosition != null) ChangedPosition(this, PosList, Ready);
			if(FinalPosition != null && Ready) FinalPosition(this, PosList, Ready);
			if(Ready) stateValue = 0;
		}

		// --- Eigenschaften -----------------------

		public int State {
			get{return stateValue;}
		}

		// --- Methoden ---------------------------------

		public void MoveHome() {
			stateValue = 1;
			for(int i = 0; i < anzMot; i++) RMotList[i].DriveHome();
		}
		public void WaitForDone() {
			for(int i = 0; i < anzMot; i++) RMotList[i].WaitForDone();
			stateValue = 0;
		}
		public void MoveTo(params int[] DestPos) {
			anzPos = DestPos.Length;
			stateValue = 1;
			for(int i = 0; i < anzPos; i++) RMotList[i].DriveTo(DestPos[i]);
		}
		public void MoveDelta(params int[] DestDelta) {
			anzPos = DestDelta.Length;
			stateValue = 1;
			for(int i = 0; i < anzPos; i++) RMotList[i].DriveDelta(DestDelta[i]);
		}

		// --- Indexer -------------------------------------------------

		public RobMotor this[int index] {
			get{return RMotList[index];}
			set{;}
		}
	}
}
